import logging
from typing import Any

from dank_mids import ENVIRONMENT_VARIABLES


class DummyLogger:
    """ Replace a `logging.Logger` object with a dummy to save precious time """
    def info(self, *args: Any, **kwargs: Any) -> None:
        ...

demo_logger = logging.getLogger("dank_mids.demo") if ENVIRONMENT_VARIABLES.DEMO_MODE else DummyLogger()




from typing import Any, Dict, Generic, Optional, TypeVar, Union

from brownie.network.contract import ContractCall, ContractTx, OverloadedMethod

_EVMType = TypeVar("_EVMType")

ContractMethod = Union[ContractCall, ContractTx, OverloadedMethod]


class _DankMethodMixin(Generic[_EVMType]):
    """A mixin class that gives brownie objects async support"""
    async def __await__(self):
        """Asynchronously call the contract method without arguments at the latest block and await the result."""
        return self.coroutine().__await__()
    async def coroutine(  # type: ignore [empty-body]
        self, 
        *args: Any, 
        block_identifier: Optional[int] = None, 
        decimals: Optional[int] = None,
        override: Optional[Dict[str, str]] = None,
    ) -> _EVMType:
        """
        Asynchronously call the contract method via dank mids and await the result.
        
        Arguments:
            - *args: The arguments for the contract method.
            - block_identifier (optional): The block at which the chain will be read. If not provided, will read the chain at latest block.
            - decimals (optional): if provided, the output will be `result / 10 ** decimals`
        
        Returns:
            - Whatever the node sends back as the output for this contract method.
        """

class DankContractCall(ContractCall, _DankMethodMixin):
    """This class serves only to enable proper type checking, you will not encounter a `DankContractCall` instance. Instead, you will have a money-patched `ContractCall`."""
class DankContractTx(ContractTx, _DankMethodMixin):
    """This class serves only to enable proper type checking, you will not encounter a `DankContractTx` instance. Instead, you will have a money-patched `ContractTx`."""
class DankOverloadedMethod(OverloadedMethod, _DankMethodMixin):
    """This class serves only to enable proper type checking, you will not encounter a `DankOverloadedMethod` instance. Instead, you will have a money-patched `OverloadedMethod`."""

DankContractMethod = Union[DankContractCall, DankContractTx, DankOverloadedMethod]
"""These classes serve only to enable proper type checking, you will not encounter a `DankContractCall`, `DankContractTx`, or `DankOverloadedMethod` instance. Instead, you will have a money-patched `ContractCall`, `ContractTx`, or `OverloadedMethod`."""


import functools
from types import MethodType
from typing import Any, Dict, Optional, Tuple, Union

from brownie import Contract
from brownie.network.contract import ContractCall, ContractTx, OverloadedMethod
from dank_mids.brownie_patch.call import _get_coroutine_fn, _skip_proc_pool
from dank_mids.brownie_patch.types import ContractMethod
from web3 import Web3


def _patch_overloaded_method(call: OverloadedMethod, w3: Web3) -> None:
    # sourcery skip: avoid-builtin-shadow
    @functools.wraps(call)
    async def coroutine(
        self: Contract,
        *args: Tuple[Any,...],
        block_identifier: Optional[Union[int, str, bytes]] = None,
        decimals: Optional[int] = None,
        override: Optional[Dict[str, str]] = None,
    ) -> Any:
        try:
            fn = self._get_fn_from_args(args)
        except ValueError as e:
            if f"Contract has more than one function '.{call._name}" in str(e) and f"You must explicitly declare which function you are calling, e.g. .{call._name}" in str(e):
                exc_str = str(e)
                breakpoint = exc_str.find("(*args)")
                raise ValueError(f"{exc_str[:breakpoint]}.coroutine{exc_str[breakpoint:]}")
            raise e

        kwargs = {"block_identifier": block_identifier, "decimals": decimals, "override": override}
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return await fn.coroutine(*args, **kwargs)

    for method in call.__dict__['methods'].values():
        if isinstance(method, (ContractCall, ContractTx)):
            method._skip_decoder_proc_pool = method._address in _skip_proc_pool
            method.coroutine = MethodType(_get_coroutine_fn(w3, len(method.abi['inputs'])), method)
    # TODO implement this properly
        #elif isinstance(call, ContractTx):
            #_patch_tx(call, w3)

    call.coroutine = MethodType(coroutine, call)




import functools
from typing import Dict, List, Optional, Union, overload

import brownie
from brownie.network.contract import (ContractCall, ContractTx, OverloadedMethod, 
                                      build_function_signature, _get_method_object)
from web3 import Web3

from dank_mids.brownie_patch.call import _patch_call
from dank_mids.brownie_patch.overloaded import _patch_overloaded_method
from dank_mids.brownie_patch.types import ContractMethod, DankContractMethod


class Contract(brownie.Contract):
    """a modified `brownie.Contract` with async and call batching functionalities"""
    @classmethod
    def from_abi(cls, *args, **kwargs) -> "Contract":
        # NOTE: just forces type checkers to work
        return super().from_abi(*args, **kwargs)
    @classmethod
    def from_ethpm(cls, *args, **kwargs) -> "Contract":
        # NOTE: just forces type checkers to work
        return super().from_ethpm(*args, **kwargs)
    @classmethod
    def from_explorer(cls, *args, **kwargs) -> "Contract":
        # NOTE: just forces type checkers to work
        return super().from_explorer(*args, **kwargs)
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # get rid of the contract call objects, we can materialize them on a jit basis
        for name in self.__method_names__:
            if name == "_name":
                # this is a property defined on _ContractBase and cannot be written to
                continue
            object.__setattr__(self, name, _ContractMethodPlaceholder)
    def __getattribute__(self, name: str) -> DankContractMethod:
        """This doesn't functionally do anythiing, it just enables type hints"""
        attr = super().__getattribute__(name)
        if attr is _ContractMethodPlaceholder:
            attr = self.__get_method_object__(name)
            object.__setattr__(self, name, attr)
        return attr
    @functools.cached_property
    def __method_names__(self) -> List[str]:
        return [i["name"] for i in self.abi if i["type"] == "function"]
    def __get_method_object__(self, name: str) -> DankContractMethod:
        from dank_mids import web3
        overloaded = self.__method_names__.count(name) > 1
        for abi in [i for i in self.abi if i["type"] == "function"]:
            if abi["name"] != name:
                continue
            full_name = f"{self._name}.{name}"
            sig = build_function_signature(abi)
            natspec: Dict = {}
            if self._build.get("natspec"):
                natspec = self._build["natspec"]["methods"].get(sig, {})

            if overloaded is False:
                fn = _get_method_object(self.address, abi, full_name, self._owner, natspec)
                _patch_call(fn, web3)
                return fn

            # special logic to handle function overloading
            elif overloaded is True:
                overloaded = OverloadedMethod(self.address, full_name, self._owner)
            overloaded._add_fn(abi, natspec)
        _patch_overloaded_method(overloaded, web3)
        return overloaded  # type: ignore [return-value]

    
@overload
def patch_contract(contract: Contract, w3: Optional[Web3] = None) -> Contract:...
@overload
def patch_contract(contract: Union[brownie.Contract, str], w3: Optional[Web3] = None) -> brownie.Contract:...
def patch_contract(contract: Union[Contract, brownie.Contract, str], w3: Optional[Web3] = None) -> Union[Contract, brownie.Contract]:
    """returns a patched version of `contract` with async and call batchings functionalities"""
    if not isinstance(contract, brownie.Contract):
        contract = brownie.Contract(contract)
    if w3 is None and brownie.network.is_connected():
        from dank_mids import dank_web3 as w3
    if w3 is None:
        raise RuntimeError("You must make sure either brownie is connected or you pass in a Web3 instance.")
    for k, v in contract.__dict__.items():
        _patch_if_method(v, w3)
    return contract

def _patch_if_method(method: ContractMethod, w3: Web3) -> None:
    if isinstance(method, (ContractCall, ContractTx)):
        _patch_call(method, w3)
    elif isinstance(method, OverloadedMethod):
        _patch_overloaded_method(method, w3)

class _ContractMethodPlaceholder:
    """A sentinel object that indicates a Contract does have a member by a specific name."""
